--
-- manureSystemHack
--
-- @author	nilBrain
-- @date	19.05.2021
-- @node	Overwritten function to change the thickness even if no mixer is present.

local thicknessText = "";

local decreaseManureThickness = function(self, superfunc, mixPerSecond, dt, updatePlane, ignoreMixer)
    if not self.isServer then
		return;
	end;
    if ignoreMixer ~= true then
		if not self.hasMixer then
			return
		end
	end;

    -- Mixed amount depends on the fill level because low fill level is mixed faster.
    local mixedAmount = (mixPerSecond / 100) / self:getFillUnitFillLevel()
    local decrease = (mixedAmount * dt) / 15
    self.thickness = math.max(self.thickness - decrease, 0)

    if updatePlane then
        self:raiseDirtyFlags(self.lagoonDirtyFlag)
    end
end

local update = function(self, dt)
    if self.isClient then
        local fillTypeName = g_fillTypeManager:getFillTypeByIndex(self:getFillUnitFillType()).title;
        g_currentMission:addExtraPrintText(thicknessText:format(fillTypeName, self.thickness * 100));
    end;
end;


local increaseManureThickness = function(self)
    if not self.isServer then
        return
    end

    -- Manure with up to 4% solids content can be handled as a liquid with irrigation equipment
    -- Manure with 4 to 10% solids content can be handled as a slurry
    local percentage = self:getFillUnitFillLevelPercentage()
    -- The more it's filled the slower thickening is.
    local mq = (1.1 - percentage) * 0.001
    if g_seasons ~= nil then
        mq = mq + (mq * 6 / g_seasons.environment.daysPerSeason)
    end

    mq = mq * 2
    self.thickness = MathUtil.clamp(self.thickness + mq, 0, 1)
    self.fillPlaneIsIdle = false
    self:raiseDirtyFlags(self.lagoonDirtyFlag)
end

if getfenv(0).FS19_manureSystem ~= nil then
	if getfenv(0).FS19_manureSystem.ManureSystemStorage ~= nil then
		getfenv(0).FS19_manureSystem.ManureSystemStorage.decreaseManureThickness = Utils.overwrittenFunction(getfenv(0).FS19_manureSystem.ManureSystemStorage.decreaseManureThickness, decreaseManureThickness);
		getfenv(0).FS19_manureSystem.ManureSystemStorage.increaseManureThickness = Utils.overwrittenFunction(getfenv(0).FS19_manureSystem.ManureSystemStorage.increaseManureThickness, increaseManureThickness);
        thicknessText = FS19_manureSystem.g_i18n:getText("info_thickness");
        getfenv(0).FS19_manureSystem.ManureSystemStorage.update = Utils.appendedFunction(getfenv(0).FS19_manureSystem.ManureSystemStorage.update, update);
	end;
end;